/*
  K-Cross Validation for TESTING a classifier.
  The results will be showed by PaneText.
*/

package racs;

import java.util.*;
import java.io.*;

import weka.core.*;
import weka.classifiers.trees.Id3;
import weka.classifiers.trees.j48.*;
import weka.classifiers.evaluation.*;
import weka.classifiers.Classifier;

public class Testing
{

 private int folds, rep;
 private Instances dataSet;
 private double [][] Accuracy;
 private int [][] Rights;
 private double [][] cAUC;
 /*
   Constructor: Inicialise the number of folds, repetitions, etc.
 */
 public Testing(int numRep, int numFolds, Instances dSet)
 {
   folds=numFolds;
   rep=numRep;
   dataSet=new Instances(dSet);
   Accuracy= new double[rep][folds];
   Rights= new int[rep][folds];
   cAUC=new double[rep][folds];
 };


 /*
   Checking RACS accuracy.
 */
 public void Checking_RACS(MetricSpace ms, int maxBranches, int classify_criterion, int splitting_criterion, boolean trace, boolean AUC, boolean Acc,int [] attributes_method)
 {
   Random random;
   int seed;
   double real_class, classified_class,right;
   Instances train, test;
   Enumeration enum;
   Instance instance;
   Id3Metric tree;
   racsUtils rUtils;

   seed=1;
   tree= new Id3Metric();
   rUtils = new racsUtils(dataSet); // setting the origanl class distribution (pre prunning)

   try
   {
     for (int i_rep = 0; i_rep < rep; i_rep++)
     {
       random = new Random(seed);
       dataSet.randomize(random); // shuffle the dataset
       //View.viewData(dataSet);
       for (int j_fold = 0; j_fold < folds; j_fold++)
       {
         // folding ...
         train = dataSet.trainCV(folds, j_fold);
         test = dataSet.testCV(folds, j_fold);
         if (trace)
         {
           System.out.println("Data for training ...");
           View.viewData(train);
         }
         // learning ...
         tree.buildClassifier(train, ms, maxBranches,splitting_criterion,attributes_method);
         if (trace) System.out.println(tree.toString());

         //  testing...
         // Accuracy evaluation
         if (Acc)
         {
           if (trace) {
             System.out.println("Data for testing ...");
             View.viewData(test);
           }
           enum = test.enumerateInstances();
           right = 0.0;
           while (enum.hasMoreElements()) {
             instance = (Instance) enum.nextElement();
             if (classify_criterion == 0)
               classified_class = tree.classifyInstance(instance, ms); // proximity criterion
             else
               classified_class = tree.classifyInstanceDensity(instance, ms); // density criterion
             real_class = instance.classValue();
             if (classified_class == real_class) {
               right++;
               if (trace)
                 System.out.println(instance + " right");
             }
             else if (trace)
               System.out.println(instance + " mistake");
           }
           Accuracy[i_rep][j_fold] = right / test.numInstances();
           Rights[i_rep][j_fold] = (int) right;
         }
         // AUC evaluation...
         if (AUC) cAUC[i_rep][j_fold]=computeAUC(tree,test);
         //System.out.println(""+cAUC[i_rep][j_fold] +"\n");
       };
     };
   } catch(Exception e) {System.err.println("Error "+e.getMessage());};
 };


 /*
   Check the accuracy of id3 method
 */
 public void Checking_ID3(boolean trace, boolean AUC, boolean Acc)
 {
   Random random;
   int seed;
   double real_class, classified_class,right;
   Instances train, test;
   Enumeration enum;
   Instance instance;
   Id3 tree;

   seed=1;
   tree = new Id3();

   try
   {
     for (int i_rep = 0; i_rep < rep; i_rep++)
     {
       random = new Random(seed);
       dataSet.randomize(random); // shuffle the dataset
       for (int j_fold = 0; j_fold < folds; j_fold++)
       {
         //folding ...
         train = dataSet.trainCV(folds, j_fold);
         test = dataSet.testCV(folds, j_fold);
         // learning ...
         tree.buildClassifier(train);
         if (trace) System.out.println(tree.toString());
         // testing...
         // Accuracy evaluation
         if (Acc)
         {
           enum = test.enumerateInstances();
           right = 0.0;
           while (enum.hasMoreElements()) {
             instance = (Instance) enum.nextElement();
             classified_class = tree.classifyInstance(instance);
             real_class = instance.classValue();
             if (classified_class == real_class) {
               right++;
               if (trace)
                 System.out.println(instance + " rigth");
             }
             else if (trace)
               System.out.println(instance + " mistake");
           };
           Accuracy[i_rep][j_fold] = right / test.numInstances();
           Rights[i_rep][j_fold] = (int) right;
         }
         // AUC evaluation...
         if (AUC) cAUC[i_rep][j_fold]=computeAUC(tree,test);
         //System.out.println(""+cAUC[i_rep][j_fold]+ "\n");
       };
     };
   } catch(Exception e) {System.err.println("Error "+e.getMessage());};
 };


 /*
   Check the accuracy of C4.5 method
 */
 public void Checking_C45(boolean trace, boolean AUC, boolean Acc)
 {
   Random random;
   int seed;
   double real_class, classified_class,right;
   Instances train, test;
   Enumeration enum;
   Instance instance;
   J48 tree;

   seed=1;
   tree = new J48();

   try
   {
     for (int i_rep = 0; i_rep < rep; i_rep++)
     {
       random = new Random(seed);
       dataSet.randomize(random); // shuffle the dataset
       for (int j_fold = 0; j_fold < folds; j_fold++)
       {
         //folding ...
         train = dataSet.trainCV(folds, j_fold);
         test = dataSet.testCV(folds, j_fold);
         // learning ...
         tree.setUnpruned(true);
         tree.buildClassifier(train);
         if (trace) System.out.println(tree.toString());
         // testing...
         // Accuracy evaluation ...
         if (Acc)
         {
           enum = test.enumerateInstances();
           right = 0.0;
           while (enum.hasMoreElements()) {
             instance = (Instance) enum.nextElement();
             classified_class = tree.classifyInstance(instance);
             real_class = instance.classValue();
             if (classified_class == real_class) {
               right++;
               if (trace)
                 System.out.println(instance + " rigth");
             }
             else if (trace)
               System.out.println(instance + " mistake");
           }
           ;
           Accuracy[i_rep][j_fold] = right / test.numInstances();
           Rights[i_rep][j_fold] = (int) right;
         }
         // AUC evaluation...
         if (AUC) cAUC[i_rep][j_fold]=computeAUC(tree,test);
       };
     };
   } catch(Exception e) {System.err.println("Error "+e.getMessage());};
 };


 /*
   Compute AUC
 */
 private double computeAUC(Classifier cl, Instances test) throws Exception
 {
   Instances tinst;
   EvaluationUtils eu;
   FastVector fpredictions;
   ThresholdCurve tc;

   tinst=null;
   eu = new EvaluationUtils();
   fpredictions = new FastVector();
   fpredictions = eu.getTestPredictions(cl, test);
   tc = new ThresholdCurve();
   tinst=tc.getCurve(fpredictions);
   return tc.getROCArea(tinst);
   //return 0.5;
 };

 /*
   Return AUC value
 */
 public double getMeanAUC()
 {
   double s=0.0;
   for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s+=cAUC[i][j];
   return s/(rep*folds);

 };

 /*
   Return AUC desviation standard value
 */
 public double getDesvAUC()
 {
   double s=0.0,aver;
   aver=getMeanAUC();
   for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s += Math.pow(cAUC[i][j] - aver, 2);
   System.out.println(""+s+"\n");
   return Math.sqrt(s/(rep*folds-1));

 };


 /*
   Return the mean accuracy.
 */
 public double getMeanAccuracy()
 {
   double s=0;
   for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s+=Accuracy[i][j];
   return s/(rep*folds);
 };

 /*
   Return the standard desviation.
 */
 public double getDesviation()
 {
   double s=0,aver;
   aver=getMeanAccuracy();
   for (int i=0;i<rep;i++) for(int j=0;j<folds;j++) s += Math.pow(Accuracy[i][j] - aver, 2);

   return Math.sqrt(s/(rep*folds-1));
 };

 /*
   Return the Right matrix: number of right classified instances in each fold and repetition
 */
 public int [][] getRightMatrix()
 {
   return Rights;
 };


 /*
   Return the Accuracy matrix: reached accuracy of each fold and repetition.
 */
  public double [][] getAccuracyMatrix()
  {
    return Accuracy;
  };

}; // end-of-class


