/*
  View items (matrices, data set, etc.) by standard output.
*/
package racs;

import weka.core.*;
import java.util.*;

// interface libraries
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.borland.jbcl.layout.*;
import javax.swing.border.*;

public class View {

 /*
   Null constructor.
 */
 public View() {};


 /*
  Print a data set by standard output.
 */
 public static void viewData(Instances data)
 {
   Enumeration enum= data.enumerateInstances();
   while (enum.hasMoreElements()) {
     Instance inst = (Instance) enum.nextElement();
     System.out.println(inst);
   }
 };


 /*
  Print a partition by standard output.
 */
 public static void viewPartition(String message,Instances[] partition)
 {
   System.out.println(message+"\n");
   for(int i=0;i<partition.length;i++)
   {
     System.out.println("Branch "+i);
     Enumeration enum= partition[i].enumerateInstances();

     while (enum.hasMoreElements())
     {
       Instance inst = (Instance) enum.nextElement();
       System.out.println(inst);
     }
   }
 };


 /*
   Print a distance matrix by standard output.
 */
 public static void viewMatrix(double [][]m)
 {
   System.out.println("");
   System.out.println("Printing matrix ...");
   for (int i=0;i<m.length;i++)
   {
     System.out.println("");
     for (int j = 0; j < m[i].length; j++) System.out.print(" " + m[i][j]);
   }
   System.out.println("");
 };


 /*
   Print all the distance matrices by standard output.
 */
 public static void viewMetricSpace(MetricSpace ms)
 {
   for (int i=0; i< (ms.getMetricSpace()).size();i++)
   {
     System.out.println("Metric Space associated to attribute "+i);
     viewMatrix((double [][]) (ms.getMetricSpace()).elementAt(i));
   }
 };


 /*
   Print the candidates, to be prototypes, by standard output.
 */
 public static void viewCandidates(String message, Candidates c )
 {
   int i;

   System.out.println("Candidates ...("+message+")\n");
   for (i=0; i< c.getNumberofCandidates();i++)
     System.out.println(c.getInstance(i)+" , "+c.getMeanDistance(i));
   System.out.println("-------------------------------------\n");
 };

 /*
  Print the a set of prototypes by standard output.
 */
 public static void viewPrototypes(String message, Instance [] protos)
 {
   System.out.println(message+"\n");
   for (int i=0;i<protos.length; i++) System.out.println(protos[i]+"\n");
 };

  /*
    View the performance of the classifier for a given fold.
    right = number of instances correctly classified.
    diference = Cardinality(fold) - right
    Acc, Mist = show or not the accuarcy and the mistakes for this fold.
    Accuracy = the accuracy reached by the classifier
  */
  public static void viewPerformance(boolean average, boolean mistakes, int [][] RightMatrix, double [][] AccuracyMatrix, JTextArea output)
  {
    int wrongs=0;
    if (average || mistakes)
    {
      for (int i = 0; i < RightMatrix.length; i++)
        for (int j = 0; j < RightMatrix[i].length; j++)
        {
          output.append("Fold " + i + " , Times " + j+"\n");
          if (mistakes)
          {
            output.append("Rights : " + RightMatrix[i][j]+"\n");
            wrongs = (int) ((RightMatrix[i][j] / AccuracyMatrix[i][j]) - RightMatrix[i][j]);
            output.append("Wrongs : " + wrongs+"\n");
          }
          if (average) output.append("Accuracy " + AccuracyMatrix[i][j]+"\n");
          output.append("--------------------------------------------------------\n");
        }
    }
  };


  /*
    View the general performance of one given classifier
  */
  public static void viewGeneralPerformance(String message,boolean AUC, double cAUC, double dcAUC, boolean Acc, double acc, double desv, JTextArea output)
  {
    String a_acc,a_desv,a_cAUC,d_cAUC;

   output.append("*********************************************\n");
   output.append(message+"\n");
   if (Acc)
   {
     a_acc = Double.toString(acc * 100).substring(0, 4);
     a_desv = Double.toString(desv * 100).substring(0, 4);
     output.append("Mean Accuracy : " + a_acc + "" + a_desv + "\n");
   };
   if (AUC)
   {
     a_cAUC = Double.toString(cAUC).substring(0, 3);
     //d_cAUC = Double.toString(dcAUC).substring(0, 3);
     output.append("AUC : " + cAUC + "" + dcAUC+ "\n");
   }
  };

}; // end-of-class
